import pandas as pd
import matplotlib.pyplot as plt

def all_against_all__individual_scatter(input_file_name, 
        output_file_stem="output"):
    '''
    For the data in a .csv file scatter plot all the data against the first 
    column, each pair on a different pair of axes, but all graphs
    next to each other on the same the same plot, 
    and then all the data against the second column, each pair on 
    a different set of axes,
    and so on. 

    Plots to .png and .pdf formats. 
    '''
    #input_file_name = "Example1_3_node_chain_NetworkValues.csv"
    #output_file_name = "Example1_AllPlots_part_"
    data_frame = pd.read_csv(input_file_name)
    outer_loop_index = 1  # for naming the different ouput files
    for (column_name,column_data) in data_frame.iteritems():
        #print(data_frame.shape[1])
        fig, axes = plt.subplots(1, data_frame.shape[1], 
            sharex=True, sharey=True)
        fig.set_facecolor("white")
        inner_loop_index = 0 
        for (comparison_column_name,comparison_column_data)  \
               in data_frame.iteritems():
            axes[inner_loop_index].scatter(column_data, \
                comparison_column_data, color="black")
            print(column_name, comparison_column_name)
            axes[inner_loop_index].set_xlabel(column_name, color="red")
            axes[inner_loop_index].set_ylabel(comparison_column_name, \
                color="red")
            axes[inner_loop_index].yaxis.set_tick_params(labelsize=15)
            axes[inner_loop_index].yaxis.set_tick_params(labelsize=15)
            
            inner_loop_index += 1
        fig.savefig(output_file_stem + str(outer_loop_index) + '.pdf')
        fig.savefig(output_file_stem + str(outer_loop_index) + '.png')
        outer_loop_index += 1

def one_plot(input_file_name, output_file_stem="output"):
    '''
    For the data in a .csv file line plot all the data against the first 
    column, all the same plot, 
    and then all the data against the second column, 
    and so on. 

    Plots to .png and .pdf formats. 
    '''
    #input_file_name = "Example1_3_node_chain_NetworkValues.csv"
    #output_file_name = "Example1_AllPlots_part_"
    data_frame = pd.read_csv(input_file_name)
    plot = data_frame.plot(xlabel = data_frame.columns[0])
    plot = data_frame.plot()
    fig = plot.get_figure()
    fig.set_facecolor("white")
    fig.savefig(output_file_stem + ".png")

def all_against_all__superposed_scatter(input_file_name, \
        output_file_stem = "output"):
    '''
    For the data in a .csv file scatter plot all the data
    against the first  column, on the same plot, and then all the data 
    against the second column, and so on. 

    Plots to .png and .pdf formats. 
    '''
    #input_file_name = "Example1_3_node_chain_NetworkValues.csv"
    #output_file_stem = "Example1_Scatter_Superposed_part_"

    print("output file name stem: ", output_file_stem)
    data_frame = pd.read_csv(input_file_name)
    outer_loop_index = 1  # for naming the different ouput files
    for (column_name,column_data) in data_frame.iteritems():
        inner_loop_index = 0 
        for (comparison_column_name,comparison_column_data) \
                in data_frame.iteritems():
            plt.scatter(column_data, comparison_column_data, 
                label = comparison_column_name)
            plt.xlabel(column_name, color="black")
            plt.legend()
            print(column_name, comparison_column_name)
            inner_loop_index += 1
        plt.savefig(output_file_stem + str(outer_loop_index)+'.pdf')
        plt.savefig(output_file_stem + str(outer_loop_index)+'.png')
        plt.show()
        outer_loop_index += 1



"""
Defines the nodes in a causal network.
"""

# This is protoype code to define nodes in a causal network, and then to
#  generate data.  The aim is to be able to explore different network 
# topologies and node relationships to allow a better understanding of 
# corellational confounding, correcting for confounding by conditioning,
# and the rules of do calculus. This is in response to reading 
# Judea Pearl and Dana Mackenzie's Book of Why,
# on the topic of causal inference.
#
# The evaluate_network_linear() call has initially been written to give a node 
# the value of the sum of all incoming nodes, scaled by their coupling
# factor, and no bias term. 
# This has been effectively replaced by evaluate_network, which allows
# the relationship between the incoming node's value and the value 
# of the node under consideration to be modelled as y = mx + c, or 
# y=mxx +c, the a bias value also to be added.
#
# The code below defines functions, classes and methods to generate 
# networks.
#
# Style conventions used:
# Class names use CamelNotation
# Method names and instance variables use lower case underscore_notation.
# Simple Instances, functions and variable names use underscore_notation.
# i.e. aiming to follow the Python pep8 style guide.
# 
 
import numpy as np
import pandas as pd
from numpy import random

def quadratic(in_value, coupling):
    return(coupling * in_value * in_value)

def linear(in_value, coupling):
    return(coupling * in_value)

def add_gaussian_noise(value, standard_deviation):
    ''' 
    returns a value from a gaussian distribution with mean being the
    "value" of the data point and standard_deviation self explanatory. 
    '''
    return random.normal(loc=value, scale=standard_deviation, size=(1))[0]

def add_noise_to_data_sd_fixed(data, noise_function, standard_deviation):
    ''' 
    uses the function passed with argument noise_function, to add
    noise onto the data set, passed as a 2d list. returns a 2d list. 
    This takes a fixed value of the standard deviation but allows a 
    variable function for adding noise. For a data set of variables 
    with different orders of magnitude it is probably better to vary 
    the standard deviation required for each variable, see e.g. 
    add_noise_to_data_sd_data_scaled().
    Data is assumed to be in the format of 
    a row of data being one set of values for the causal network.
    noise_function  currently only has the choice of 
    add_gaussian_noise(value, standard_deviation)

    could refactor the code to turn the data set into a collection of 
    network instances, improving the object oriented nature of the code. 
    '''
    data_list = []
    data_list_row = []
    for row in data:
        for value in row:        
            data_list_row.append(noise_function(value, standard_deviation))
        data_list.append(data_list_row)
        data_list_row = []
    return data_list

def add_noise_to_data_sd_data_scaled2(data, standard_deviation_scale):
    ''' 
    adds Gaussian noise to the data, passed as a 2d list, 
    where the standard deviation of the noise is scaled by 
    the mean of the data column. Returns a numpy array. 
    Data is assumed to be in the format of 
    a row of data being one set of values for the causal network.
    the  currently only has the choice of 
    add_gaussian_noise(value, standard_deviation)

    This implementation uses slicing to select the appropriate data, 
    c.f. add_noise_to_data_sd_data_scaled, which does the same but with
    whole array manipulation.
    could refactor the code to turn the data set into a collection of 
    network instances, improving the object oriented nature of the code. 
    '''
    data_array = np.array(data)
    for column in range(data_array.shape[1]):
        column_mean =np.mean(data_array[:,column:column+1])
        standard_deviation = standard_deviation_scale * column_mean
        noised_data = data_array[:,column:column+1].copy() + np.random.normal(loc=0,
            scale=standard_deviation, size=(data_array.shape[0],1))
        data_array[:,column:column+1] = noised_data.copy()
    return data_array

def add_noise_to_data_sd_data_scaled(data, standard_deviation_scale):
    ''' 
    adds Gaussian noise to the data, passed as a 2d list, 
    where the standard deviation of the noise is scaled by 
    the mean of the data column. Returns a numpy array. 
    Data is assumed to be in the format of 
    a row of data being one set of values for the causal network.
    Currently only adds gaussian noise via: 
    add_gaussian_noise(value, standard_deviation)

    could refactor the code to turn the data set into a collection of 
    network instances, improving the object oriented nature of the code. 
    '''
    data_array = np.array(data)
    column_means = np.mean(data_array, axis=0)
    standard_deviation = column_means * standard_deviation_scale
    noise_data = np.random.normal(loc=0, 
            scale=standard_deviation, size= data_array.shape)
    data_array = data_array + noise_data.copy()
    return data_array    

class ConnectionOverwriteError(Exception):
    def __init__(self):
        super().__init__()
   
    def __str__(self):
        return "Trying to set a node that already exists. \n \
                   Try revalue_out_connection()"

class Node:
    """
    Create a node object with a node value, name, dictionary of input
    connections, dictionary of output connections.
    """

    def __init__(self, name, bias=0):
        """ 
        Initialise the node name, value, and connection lists. 
        
        node_previous_value allows comparison with node_value to see 
        if convergence of node value has been achieved when iteratively
        updating the network. 
        """
        self.node_value = bias
        self.node_previous_value = bias
        self.node_bias = bias # value of node with no inputs
        self.name = name
        self.out_connections = {}
        self.in_connections = {}

    def add_out_connection(self, node, coupling_data):
        """ 
        Add entry to the node's out_connections dictionary  
        and update the corresponding in_connections dictionary 
        of the node connecting to self . 

        The dictionary keys are Node objects, so can be iterated over
        to return the Nodes stored in the in_connections and 
        out_connections dictionaries.

        node -- the node to connect to the object node, i.e.  recorded
        in its in_connections dictionary.
        coupling_data -- could be a value for the coupling between the 
        two nodes, for use with evaluate_function(), which assumes a 
        linear coupling between the node values
        (i.e. first order polynomial).  
        Or a tuple with (coupling_value, coupling_function), where 
        coupling_function is a function defining the nature of the 
        coupling between nodes, linear or quadratic, at the time of 
        writing this update (06/06/2022), for use with 
        evaluate_network()   

        coupling_function = node.in_connections[in_connection][1]
        incoming_value = in_connection.node_value 
        coupling_value = node.in_connections[in_connection][0]
        """

        if node in self.out_connections:
            raise ConnectionOverwriteError()
        # Add to dictionary of couplings of this node (self).
        self.out_connections[node] = coupling_data
        # Add in connection to receiving node (node)
        node.in_connections[self] = coupling_data


    def revalue_out_connection(self, node, coupling_data):
        """
        Changes data of an existing node. c.f. add_out_connection()
        """
        self.out_connections[node] = coupling_data
        node.in_connections[self] = coupling_data
         


    def add_in_connection(self, node, coupling_data):
        """ 
        Add entry to the node's in_connections dictionary.  
        Also update the corresponding out_connections dictionary 
        of the node connecting to self. 

        Not really needed if add_out_connection is used, since they
        effectively do the same from different perspectives. 

        A longer description is found in add_out_connection(), which 
        uses the same syntax/structures. 
        """
        self.in_connections[node] = coupling_data
        # Adds an out connection to the giving node
        # - it is inefficient to write in and out arrows explicitly,
        # since it duplicates, but should make user experience easier
        node.in_connections[self] = coupling_data

    def set_value(self, value):
        self.node_value = value

    def set_previous_value(self, value):
        self.node_previous_value = value

    def set_bias(self, bias):
        self.node_bias = bias

    def random_real_value_bias(self, min, max, seed):
        '''
        reassigns the node with random
        real number value. 

        uses python random module of the standard python package 
        (node object, min value of node, max value)
        '''
        random.seed(seed)
        value = min + random.random() * (max - min)
        self.set_bias(value)
            
    def random_int_value_bias(self, min, max, seed):
        '''
        reassigns the nodes defined in list_of_node_range_tuples with random
        integer number values. 

        uses python random module of the standard python package
        '''
        random.seed(seed)
        value = random.randint(min,max)
        self.set_bias(value)

    def print_node_data(self):
        """
        Print the name, value, out connections, in connections of
        a node.
        """
        print("\n Start node data \n")
        print("name")
        print(self.name)
        print("value ")
        print(self.node_value)
        print("-----------------------\n")
        print("Out connections \n")
        for connected_node in self.out_connections.keys():
            print(connected_node.name) # dictionary key is a node object
            print(self.out_connections[connected_node])

        print("-----------------------\n")
        print("In connections \n")
        for connected_node in self.in_connections.keys():
            print(connected_node.name)
            print(self.in_connections[connected_node])
        print("Done node data \n")




class NetworkInstance:
    """
    Create a network object from a list of nodes defined by the
    Node class.
    """

    def __init__(self, node_list):
        self.network_node_list = node_list

    def evaluate_network_linear(self):
        """
        Iterate over nodes, once, and calculate a value for each node, 
        assuming a y=mx relationship.

        Only traverses the network once, so may not be self consistent.
        Convergence of the network is achieved by calling
        evaluate_network multiple times and checking for convergence,
        e.g. as in converge_network.
        """
        for node in self.network_node_list:
            value = 0
            #print(node.name, node.node_value)
            if not node.in_connections:# Start node  not dependent on an 
                                       # input does not need updating
                node.node_value = node.node_bias
                continue
            else: # needs updating according to incoming connections
                # for each node, for each input connection evaluate its
                # contribution to the node value.  
                # node.in_connections[in_connection] stores the 
                # coupling value of the in coming node. I.e. the 
                # multiplier of the incoming node's value. 
                for in_connection in node.in_connections.keys():
                    value = value + (in_connection.node_value  
                                    * node.in_connections[in_connection]) 
            
            value = value + node.node_bias
            node.set_value(value)


    def evaluate_network(self):
        """
        Iterate over nodes, once, to calculate a value for each node.
        
        Only traverses the network once, so may not be self consistent.
         Convergence of the network is achieved by calling
        evaluate_network multiple times and checking for convergence,
        e.g. as in converge_network.
        
        Requires coupling_data, defined by  add_out_connection() or 
        add_in_connection(), and stored in the in_connections and
        out_connections dictionaries under the node key.  coupling_data
        is a tuple consisting of (coupling_value, coupling_function), 
        but could easily be extended to contain further data in the 
        future. 

        """
        for node in self.network_node_list:#for each node in the network
            value = 0
            #print(node.name, node.node_value)
            if not node.in_connections:# Start node  not dependent on an 
                                       # input; does not need updating
                node.node_value = node.node_bias #Ensure bias and value
                                                 #equal when iterating value
                continue
            else:# needs updating according to incoming connections
                # for each node (current loop), for each input 
                # connection evaluate its
                # contribution to the node value.  
                # node.in_connections[in_connection] stores the 
                # coupling value of the in coming node. I.e. the 
                # multiplier of the incoming node's value. 
                # coupling_function is linear or quadratic or etc. 
                for in_connection in node.in_connections.keys():
                    #print("connection data: ", 
                    #     node.in_connections[in_connection])
                    coupling_function = node.in_connections[in_connection][1]
                    incoming_value = in_connection.node_value 
                    coupling_value = node.in_connections[in_connection][0]
                    value = (
                        value 
                        + coupling_function(incoming_value, coupling_value)
                    )
                    
            value = value + node.node_bias
            node.set_value(value)           

    def converge_network(self):
        """
        Use the evalute_network method to calculate the value
        at every node and then repeat until a consistent value is
        obtained.

        The value at each node depends on the value at the nodes
        connected to it, so the network may need to be traversed
        several times before it is self consistent.
        """
        # Check value at all nodes.
        # Re-evaluate.
        # Check if different from previous evaluation. If so repeat, else
        # Return.
        for node in self.network_node_list:
            node.set_previous_value(node.node_value)
        self.evaluate_network()
        for node in self.network_node_list:
            # If a node is found to differ from it's previous value
            # we need to save the current values as previous
            # and iterate the values of nodes in the network again
            while abs(node.node_previous_value - node.node_value) > 0.01:
                node.set_previous_value(node.node_value)
                print("re-evaluate")
                self.evaluate_network() 



    def converge_network_linear(self):
        """
        Use the evalute_network method to calculate the value
        at every node and then repeat until a consistent value is
        obtained.

        The value at each node depends on the value at the nodes
        connected to it, so the network may need to be traversed
        several times before it is self consistent.
        """
        # Check value at all nodes.
        # Re-evaluate.
        # Check if different from previous evaluation. If so repeat, else
        # Return.
        for node in self.network_node_list:
            node.set_previous_value(node.node_value)
        self.evaluate_network_linear()
        for node in self.network_node_list:
            # If a node is found to differ from it's previous value
            # we need to save the current values as previous
            # and iterate the values of nodes in the network again
            while abs(node.node_previous_value - node.node_value) > 0.01:
                node.set_previous_value(node.node_value)
                print("re-evaluate")
                self.evaluate_network_linear() 

    def print_network(self):
        for node in self.node_list:
            # print(node.Name)
            node.print_node_data()


    def scan_parameters(self, node_and_range):
        variable_node = node_and_range[0]
        range = node_and_range[1]
        data_list = []
        for value in range:
            variable_node.set_bias(value) # for scan, bias and value
                                          # should be held the same.
                                          # evaluate_network() sets 
                                          # value to bias
                                          # for orphan nodes that drive
                                          # a network.
            self.evaluate_network()
            self.converge_network()
            data_list_row = []
            for node in self.network_node_list:
                data_list_row.append(node.node_value)
            data_list.append(data_list_row)
            data_list_row = []
        return data_list

    def scan_parameters2D(self, node_and_range):
        variable_node1 = node_and_range[0]
        range1 = node_and_range[1]
        variable_node2 = node_and_range[2]
        range2 = node_and_range[3]
        data_list = []
        for value1 in range1:
            variable_node1.set_bias(value1)# for scan, bias and value
                                          # should be held the same.
                                          # evaluate_network() sets 
                                          # value to bias
                                          # for orphan nodes that drive
                                          # a network.
            for value2 in range2:
                variable_node2.set_bias(value2)
                self.evaluate_network()
                self.converge_network()
                data_list_row = []
                for node in self.network_node_list:
                    data_list_row.append(node.node_value)
                data_list.append(data_list_row)
                data_list_row = []
        return data_list    

    def scan_parameters_linear(self, node_and_range):
        variable_node = node_and_range[0]
        range = node_and_range[1]
        data_list = []  # to store output from all iterations
        data_list_row = []  # to store output from one iteration
        for value in range:
            variable_node.set_bias(value) # when scanning, bias and value
                                          # should be held the same.
                                          # evaluate_network sets value 
                                          # to bias
                                          # for orphan nodes that drive
                                          # a network.
            self.evaluate_network_linear()
            self.converge_network_linear()
            for node in self.network_node_list:
                data_list_row.append(node.node_value)
            data_list.append(data_list_row)
            data_list_row = []
        return data_list

    def write_network_parameters(self, filename):
        """ Write network parameters to file. I.e. connections, 
        couplings, values at nodes."""
        #with open(filename, 'w') as filepointer:
        filepointer = open(filename, "w")
        data_list = [
            "NodeName",
            "CurrentValue",
            "InComingConnectionFromNodes",
            "InComingValues"]
        filepointer.write(str(data_list) + " \n")
        for node in self.network_node_list:
            #print(node.Name, node.NodeValue)
            data_list = [node.name, node.node_bias, node.node_value]
            #network_dataframe.append(pd.Series(DataList, index=network_dataframe.columns), ignore_index=True)
            for in_connection in node.in_connections:
                # in_connection is a dictionary key but that key is a 
                # node object.
                data_list.append(in_connection.name)
                data_list.append(node.in_connections[in_connection])
            # print(network_dataframe)
            filepointer.write(str(data_list) + " \n")
        # print(network_dataframe)
        filepointer.close()

    def make_network_parameters_dataframe(self):
        """ Write network parameters to file. I.e. connections, 
        couplings, values at nodes."""
        data_list = []
        for node in self.network_node_list:
            data_list_row = [node.name, node.node_bias, node.node_value]
            for in_connection in node.in_connections:
                if in_connection.name:
                    # in_connection is a dictionary key but that key is 
                    # a node object.
                    data_list_row.append(in_connection.name)
                    data_list_row.append(node.in_connections[in_connection])
            data_list.append(data_list_row)
        # This section to sort out column names. Could be tidied up by 
        # considering the dimension of the rows and adjusting the 
        # headings dynamically.
        # column_names = 
        #  ["NodeName", "CurrentValue", 
        #   "InComingConnectionFromNodes", "InComingValues"]
        network_dataframe = pd.DataFrame(data_list)
        network_dataframe.rename(
            columns={
                0: 'NodeName',
                1: 'Node Bias Value',
                2: 'CurrentValue',
                3: 'InComingConnectionFromNodes',
                4: 'InComingValues - Coupling and FunctionForm'},
            inplace=True)
        return network_dataframe

    def list_of_node_names(self):
        column_names = []
        for node_entry in self.network_node_list:
            column_names.append(node_entry.name)
        return column_names
    











'''
Calculate multiple regression, every variable described as a 
function of all other variables.
Iterate column by column for every column against 
Calculation results returned to stdout. 
'''

from sklearn import linear_model
import pandas as pd

def multi_regress_from_csv(filename):
    '''Regress one column against all other columns in a csv file, 
    iterating to include every column in turn as the dependent variable'''
    data_frame = pd.read_csv(filename)
    multi_regress_from_dataframe(data_frame)

def multi_regress_from_dataframe(dataframe):
    '''Regress one column against all other columns in a csv file, 
    iterating to include every column in turn as the dependent variable'''
    data_frame = dataframe
    slope_list = []
    intercept_list = []
    correlation_list = []
    p_list = []
    std_err_list = []
    key_list = []
    data_names = []
    
    for (column_name,column_data) in data_frame.iteritems():
        dependent_data = column_data
        dependent_name = column_name
        data_names.append(dependent_name)
        slope_row_data = []
        key_row_data = []
        intercept_row_data = []
        correlation_row_data = []
        p_row_data = []
        std_err_row_data = []

        independent_data = data_frame.loc[:, data_frame.columns != column_name]
        print("dependent name: ", dependent_name)
        independent_names = list(data_frame.loc[:, data_frame.columns != column_name])
        print("independent names: ", independent_names)
        # with sklearn
        regression_model = linear_model.LinearRegression()
        regression_model.fit(independent_data, dependent_data)

        print('Intercept: \n', regression_model.intercept_)
        print('Coefficients: \n', regression_model.coef_)



'''
code to read a .csv and for each column in turn perform a multiple 
regression with respect to all the other columns. Uses random 
initialisation and steepest gradient descent.
The code works but needs refactoring 
'''

import pandas as pd
import numpy as np

def multi_regress_from_csv_2(filename, threshold=0.001, random = 1, \
    max_iter = 0, learning_rate = 0.01):
    '''
    Regresses one column against all other columns in a csv file, 
    iterating to include every column in turn as the dependent variable.
    max_iter = 0 implies no restriction on number of iterations, otherwise
    the value is the maximum number of gradient descent steps.
    random = 1 - use random starting values of the parameters
    random = 0 - use the average gradient as starting parameter values, 
    intercept 0.
    Threshold, value of mean squared error below which to stop optimising
    '''
    data_frame = pd.read_csv(filename)
    
    for (column_name, column_data) in data_frame.iteritems():
        dependent_data = column_data
        dependent_name = column_name
        # independent data is every column except the one selected by this 
        # iteration of the for loop.
        independent_data = data_frame.loc[:, data_frame.columns != column_name] 
        independent_names = list(data_frame.loc[:, data_frame.columns != \
            column_name])
        #print("dependent name: ", dependent_name, "independent names: ", \
        #    independent_names)
        print(f"Calling Regression with threshold = {threshold}, random = \
            {random}, max_iter = {max_iter}")
        regression_out_parameters = regress(independent_data, dependent_data, \
            threshold, random, max_iter, learning_rate)
        r_squared = calculate_r_squared(regression_out_parameters, independent_data, dependent_data)
        #print("r_squared", r_squared)
        # Get create a numpy array with names identifying the parameters.
        # Insert new column name to indentify the intercept
        independent_names.insert(0, "intercept") 
        independent_names = np.array(independent_names)
        independent_names = np.expand_dims((independent_names), axis=1)
        independent_names = np.transpose(independent_names)
        # columns names to the parameter array - except transpose so to rows.
        regression_out_parameters = np.transpose(regression_out_parameters)
        regression_out_parameters = np.append(independent_names,regression_out_parameters, axis=0)
        regression_out_parameters = np.append(regression_out_parameters, [["r_squared "], [r_squared]], axis=1)
        # save to final store so don't overwrite on next iteration
        if  "all_regression_out_parameters" in locals():
            all_regression_out_parameters = np.append(all_regression_out_parameters, np.array([["  ", "  ",]]) ,axis = 0)
            all_regression_out_parameters = np.append(all_regression_out_parameters, np.array([["dependent", dependent_name]]) ,axis = 0)
        else:
            all_regression_out_parameters = np.array([["  ", "  "]])
            all_regression_out_parameters = np.append(all_regression_out_parameters, np.array([["dependent", dependent_name]]) ,axis = 0)
        all_regression_out_parameters = np.append(all_regression_out_parameters,  np.transpose(regression_out_parameters), axis=0)      
        #print("all_regression_out_parameters", all_regression_out_parameters)
    return all_regression_out_parameters

def calculate_r_squared(regression_parameters,independent_data,dependent_data):
    number_of_training_examples = independent_data.shape[0]
    independent_data = np.append(np.ones((number_of_training_examples,1)), \
        independent_data, axis=1)
    Y = np.expand_dims(dependent_data, axis=1)
    Y_hat = np.dot(independent_data, regression_parameters)
    sum_squares_residuals = mse(Y, Y_hat, number_of_training_examples)
    total_sum_squares = np.var(Y)
    r_squared = 1 - (sum_squares_residuals/total_sum_squares)
    return r_squared

# loss function aka error function:
def mse(y, y_hat, number_of_training_examples): 
  '''Calculate the mean squared error (mse) loss: (y_hat - y)squared. 
  y_hat is the estimate from the network, 
  y is the ground truth.'''
  return np.sum(np.square(y_hat - y), axis=0)/number_of_training_examples

def mse_gradient(Y, X, Y_hat, number_of_training_examples):
  '''gradient of mse: mean squared error loss: (y_hat - y). y_hat 
  is the estimate from the network, 
  y is the ground truth.'''
  return 0.5*np.sum(np.multiply((Y_hat - Y),X), axis=0)/number_of_training_examples

def regress(independent_data, dependent_data, threshold=0.9, random = 1, max_iter = 0, learning_rate = 0.005):
    #add column of ones, for the intercept, shape independent_data.shape[0],1
    #print("in regress function")
    column_length = independent_data.shape[0]
    independent_data = np.append(np.ones((column_length,1)), independent_data, axis=1)
    Y = np.expand_dims(dependent_data, axis=1)
    parameter_count = np.shape(independent_data[1])[0]
    #print("parameter count", parameter_count)
    
    # random initialisation or comparison to mean gradient and set intercept 0.
    print("random, ", random) 
    if random == 1:
        print("using pseudo random initial parameter values.")
        parameters = np.random.random((parameter_count,1))
    elif random == 0:
        print("setting initial parameters using gradient estimates.")
        grad = []
        for grad_count in range(parameter_count):
            # estimate the gradients
            delta_dependent = np.max(dependent_data) - np.min(dependent_data)
            grad.append((np.max(independent_data[grad_count]) \
                 - np.min(independent_data[grad_count])) /delta_dependent)
            grad_count += 1  
            #print("grad_count, ", grad_count)
        parameters = np.append(np.array([0.0]),np.array(grad[1:]))
        Y = np.expand_dims(dependent_data, axis=1)
        parameters = np.expand_dims(parameters, axis = 1)
        #print("parameters, ", parameters)
    else:
        print("random should be 1 or 0. ", random, " is not a valid value.")
        exit()

    Y_hat = np.dot(independent_data, parameters)
    number_of_training_examples = Y.shape[0]
    error = mse(Y, Y_hat, number_of_training_examples)
    #print("mean squared error ", error)
    iterations = 0
    while  error > threshold:
        gradient = np.expand_dims(mse_gradient(Y, independent_data, Y_hat, \
            number_of_training_examples), axis=1)
        #print("gradient: \n", gradient) 
        parameters = parameters - learning_rate * gradient
        #print("parameters: \n", parameters)
        Y_hat = np.dot(independent_data, parameters)
        #print("Y_hat", Y_hat)
        error = mse(Y, Y_hat, number_of_training_examples)
        #print("mean squared error: ", error)
        if max_iter > 0:
            if iterations == max_iter:  break
        iterations += 1
        #print("End of iteration: ", iterations, " of ", max_iter, " iterations")
        
    return parameters





'''
calculate pairwise regression analysis comparing every pair of columns.
Iterate column by column for every column against 
every other column in a csv file and save in a dataframe.
Dataframe written to .tex, .csv, .html files. 
'''

from scipy import stats
import pandas as pd

def regress_from_csv(filename):
    '''All by all regression of the columns in a csv file.'''
    data_frame = pd.read_csv(filename)
    # initialise data storage lists
    slope_list = []
    intercept_list = []
    correlation_list = []
    p_list = []
    std_err_list = []
    key_list = []
    data_names = []

    # nested for loops for all against all comparison
    for (column_name,column_data) in data_frame.iteritems():
        dependent_data = column_data
        dependent_name = column_name
        data_names.append(dependent_name)
        slope_row_data = []
        key_row_data = []
        intercept_row_data = []
        correlation_row_data = []
        p_row_data = []
        std_err_row_data = []
        # if stdout_flag == true:
        # print to stdout stuff picked up in the next loop - 
        # next line provides the header for the output to stdout
        print("calculation_key, slope, intercept, r_value, p_value, std_err")
        for (comparison_column_name,comparison_column_data) in data_frame.iteritems():
            independent_data = comparison_column_data
            independent_name = comparison_column_name
            # scipy regression
            slope, intercept, r_value, p_value, std_err = \
                stats.linregress(independent_data, dependent_data)

            # if stdout_flag == true:
            # print to stdout stuff.
            calculation_key =  independent_name + dependent_name
            print(calculation_key, slope, intercept, r_value, p_value, std_err)
            
            #add to their row list for each output, slope, intercept ...
            slope_row_data.append(slope)
            key_row_data.append(calculation_key)
            intercept_row_data.append(intercept)
            correlation_row_data.append(r_value)
            p_row_data.append(p_value)
            std_err_row_data.append(std_err)
        # add row data to the full list for each slope, intercept etc. 
        slope_list.append(slope_row_data)
        key_list.append(key_row_data)
        intercept_list.append(intercept_row_data)
        correlation_list.append(correlation_row_data)
        p_list.append(p_row_data)
        std_err_list.append(std_err_row_data)
    # convert  to dataframes
    slope_dataframe = pd.DataFrame(slope_list)
    intercept_dataframe = pd.DataFrame(intercept_list)
    correlation_dataframe = pd.DataFrame(correlation_list)
    p_dataframe = pd.DataFrame(p_list)
    std_err_dataframe = pd.DataFrame(std_err_list)

    # Label columns/rows as Inpendent/Dependent variables.
    for dataframe in [slope_dataframe, intercept_dataframe, 
                     correlation_dataframe, p_dataframe, std_err_dataframe]:
        dataframe.columns = pd.MultiIndex.from_product([["Dependent"],
                                                       data_names])
        dataframe.index = pd.MultiIndex.from_product([["Independent"], 
                                                     data_names])
    return slope_dataframe, intercept_dataframe, correlation_dataframe, \
        p_dataframe, std_err_dataframe

def dataframe_to_html(dataframe, filename_stem):
    '''write dataframe to html file called filename_stem".html"'''
    html_filepointer = open(filename_stem+".html", "w")
    html_filepointer.write(dataframe.to_html())
    html_filepointer.close()

def dataframe_to_tex(dataframe, filename_stem):
    '''write dataframe to tex file called filename_stem".tex"''' 
    html_filepointer = open(filename_stem+".tex", "w")
    html_filepointer.write(dataframe.to_latex())
    html_filepointer.close()
        
def dataframe_to_csv(dataframe, filename_stem):
    '''write dataframe to .csv file called filename_stem".csv"''' 
    html_filepointer = open(filename_stem+".csv", "w")
    html_filepointer.write(dataframe.to_csv())
    html_filepointer.close()




import pandas as pd
import sys, os
sys.path.append('..\\..\\') # to pick up network_generator.py
from network_generator import *

def scan_1D_of_network_add_noise(node_list, variable_to_scan, scan_range_lower,
        scan_range_upper, filestem):
    '''
    scans one variable of the network between the values scan_range_lower and
    scan_range_upper. Adds noise to the converged data set and writes out the
    clean and noisy data to two different files, aswell as files with network
    parameters.
    '''
    network = NetworkInstance(node_list)
    network.evaluate_network_linear()
    network.converge_network_linear()
    network_dataframe = network.make_network_parameters_dataframe()
    print(network_dataframe)

    ## 2.1 block to iterate through values of X 
    node_and_range = (variable_to_scan, range(scan_range_lower, scan_range_upper)) 
    # node_and_range defined as a tuple with the idea that a list of tuples 
    #could be passed to scan_parameters, to scan multiple parameters at once.
    data_from_scanningX = network.scan_parameters_linear(node_and_range)

    ## 2.2 write scanned data to file
    column_names = network.list_of_node_names()  
    network1_values_dataframe = pd.DataFrame(data_from_scanningX, columns=column_names)
    network1_values_dataframe.to_csv(
                                    filestem+"_NetworkValues.csv",
                                    index=False,
                                    na_rep='None')

    # Write network parameters to file, including last converged values. 
    network.write_network_parameters(filestem+"_NetworkParameters.txt")  
    network_parameters_frame=network.make_network_parameters_dataframe()
    network_parameters_frame.to_csv(
                                    filestem+"_NetworkParameters.csv",
                                    index=False,
                                    na_rep='None')

    ## 2.3 add noise.
    ##

    # Add noise to the data. We're going to do that three times here,  
    # too few to bother with a loop, to increase the data set size.
    standard_deviation_scale = 0.1 # standard deviation of gaussian noise
                                # will be standard_deviation_scale * mean 
                                # of data column.

    data_array = add_noise_to_data_sd_data_scaled(data_from_scanningX, standard_deviation_scale)
    data_array = np.concatenate([data_array, 
        add_noise_to_data_sd_data_scaled(data_from_scanningX, standard_deviation_scale)], axis = 0 )
    data_array = np.concatenate([data_array, 
        add_noise_to_data_sd_data_scaled(data_from_scanningX, standard_deviation_scale)], axis = 0 )

    # write scanned data to file
    network2_values_dataframe = \
        pd.DataFrame(data_array, columns=column_names)
    network2_values_dataframe.to_csv(
                                    filestem+"_NetworkValues_noisy.csv",
                                    index=False,
                                    na_rep='None')


from setuptools import setup

from setuptools import setup
setup(name='causaln',
version='1.0',
description='for defining causal networks and generating example data',
url='https://github.com/PeterJamesWinn',
author='Peter James Winn',
license='MIT',
packages=['causaln'],
zip_safe=False)
